%% Example
% In this example we show all optional options for compositePebiGrid.

%{
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Copyright (C) 2016 Runar Lie Berge. See COPYRIGHT.TXT for details.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%}  
%% wellLines
% This sets the wells in our reservoir. The wells are stored as a cell
% array, each element corresponding to one well
w = {[0.2,0.8;0.5,0.7;0.8,0.8],...
     [0.5,0.2]};
gS = [0.1,0.1];
pdims=[1,1];
G = compositePebiGrid(gS, pdims,'wellLines',w);

plotGrid(G);
hold on
plotGrid(G,G.cells.tag,'facecolor','b')
axis equal
%% wellGridFactor
% The wellGridFactor sets the relative size of the well cells. If
% wellGridFactor=0.5 the well cells will have about half the size of the
% reservoir sites:
w = {[0.2,0.3;0.8,0.7]};
gS = [0.1,0.1];
pdims=[1,1];
G1 = compositePebiGrid(gS, pdims,'wellLines',w,'wellGridFactor',1);
G2 = compositePebiGrid(gS, pdims,'wellLines',w,'wellGridFactor',1/2);

figure()
plotGrid(G1)
title('wellGridFactor=1')
axis equal
figure()
plotGrid(G2)
title('wellGridFactor=1/2')
axis equal


%% mlqtMaxLevel
% mlqtMaxLevel sets how many steps of local grid refinement we create
% around each well.

w = {[0.2,0.3;0.8,0.7]};
gS = [0.1,0.1];
pdims=[1,1];
G1 = compositePebiGrid(gS, pdims,'wellLines',w,'wellGridFactor',1/2,'mlqtMaxLevel',1);
G2 = compositePebiGrid(gS, pdims,'wellLines',w,'wellGridFactor',1/4,'mlqtMaxLevel',2);

figure()
plotGrid(G1)
title('mlqtMaxLevel=1')
axis equal
figure()
plotGrid(G2)
title('mlqtMaxLevel=2')
axis equal

%% mlqtLevelSteps
% We can define which distance from a well triggers which level of
% mlqt-refinement. 
w = {[0.2,0.3;0.8,0.7]};
gS = [0.1,0.1];
pdims=[1,1];
lev1 = [0.05,0.1];
lev2 = [0.1,0.2];
G1 = compositePebiGrid(gS, pdims,'wellLines',w,'wellGridFactor',1/4, ...
                      'mlqtMaxLevel',2,'mlqtLevelSteps',lev1);
G2 = compositePebiGrid(gS, pdims,'wellLines',w,'wellGridFactor',1/4, ...
                      'mlqtMaxLevel',2,'mlqtLevelSteps',lev2);

figure()
plotGrid(G1)
title('mlqtLevelSteps=[0.05,0.1]')
axis equal
figure()
plotGrid(G2)
title('mlqtLevelSteps=[0.1,0.2]');
axis equal

%% wellRho
% A function that sets the relative size of the fault sites in the domain.
w = {[0.2,0.3;0.8,0.7]};
gS = [0.1,0.1];
pdims=[1,1];
wellRho = @(p) 1 - 0.9*p(:,1);
G = compositePebiGrid(gS, pdims,'wellLines',w,'wellRho',wellRho);

figure()
plotGrid(G)
plotGrid(G,G.cells.tag,'facecolor','b')
title('wellRhoRho=@(p) 1 - 0.9*p(:,1)')
axis equal
%% ProtLayer
% Adds a protection layer around wells. The protection sites are place
% normal along the well path
x = linspace(0.2,0.8);
y = 0.5+0.1*sin(pi*x);
w = {[x',y']};
gS = [0.1,0.1];
pdims=[1,1];
G = compositePebiGrid(gS, pdims,'wellLines',w,'protLayer',true);

figure()
plotGrid(G);
hold on
plotGrid(G,G.cells.tag,'facecolor','b')
title('Protection Layer on')
axis equal

%% protD
% This parameter sets the distance from the protection sites to the well
% path. This is cellarray of functions, one for each well path.
w = {[0.2,0.8;0.8,0.8],...
     [0.5,0.2;0.1,0.5]};
gS = [0.08,0.08];
pdims=[1,1];
protD = {@(p)0.01+0.1*p(:,1), @(p) 0.01*ones(size(p,1),1)};
G = compositePebiGrid(gS, pdims,'wellLines',w,'protLayer',true,'protD',protD);

figure()
plotGrid(G);
hold on
plotGrid(G,G.cells.tag,'facecolor','b')
title('Specifying Protection Distance')
axis equal



%% faultLines
% This sets the faults in our reservoir. The wells are stored as a cell
% array, each element corresponding to one fault
f = {[0.2,0.8;0.5,0.7;0.8,0.8],...
     [0.5,0.2;0.1,0.5]};
gS = [0.1,0.1];
pdims=[1,1];
figure()
G = compositePebiGrid(gS, pdims,'faultLines',f);
plotGrid(G);
hold on
plotFaces(G,G.faces.tag,'edgeColor','r')
title('Fault Lines')
axis equal


%% faultGridFactor
% The faultGridFactor sets the relative distance between the fault cells 
% along the fault paths. If fautlGridFactor=0.5 the fault cells will be 
% have spaceing about half the size of the reservoir cells:
f = {[0.2,0.3;0.8,0.7]};
gS = [0.1,0.1];
pdims=[1,1];
G1 = compositePebiGrid(gS, pdims,'faultLines',f,'faultGridFactor',1);
G2 = compositePebiGrid(gS, pdims,'faultLines',f,'faultGridFactor',1/2);

figure()
plotGrid(G1)
title('faultGridFactor=1')
axis equal
figure()
plotGrid(G2)
title('faultGridFactor=1/2')
axis equal

%% circleFactor
% The fault sites are generated by setting placing a set of circles along
% the fault path with a distance gS*faultGridFactor. The radius of the
% circles are gS*faultGridFactor*circleFactor. The fault sites are place
% at the intersection of these circles.
f = {[0.2,0.3;0.8,0.7]};
gS = [0.1,0.1];
pdims=[1,1];
G1 = compositePebiGrid(gS, pdims,'faultLines',f,'circleFactor',0.55);
G2 = compositePebiGrid(gS, pdims,'faultLines',f,'circleFactor',0.9);

figure()
plotGrid(G1)
title('circleFactor=0.55')
axis equal
figure()
plotGrid(G2)
title('circleFactor=0.9')
axis equal

%% PolyBdr
% Create a non-square reservoir domain
bdr   = [0,0;-0.5,1;0.5,1.5;1,1];
gs    = [0.1,0.1];
pdims = [1,1]; % can be set to anything
G = compositePebiGrid(gs,pdims,'polyBdr',bdr);

figure()
plotGrid(G)
title('Polygon boundary')
axis equal tight
